import os
import shutil
from typing import Iterator, Tuple, Union, List

import requests
from filelock import FileLock
from hbutils.collection import nested_map
from hbutils.string import format_tree

from .base import _soft_name_strip
from .model import ModelManager
from ..client import find_model, OFFLINE_MODE, OfflineModeEnabled


class LocalModelNotFound(Exception):
    pass


class LocalModelDuplicated(Exception):
    pass


class DispatchManager:
    """
    Management of all models.
    """

    def __init__(self, root_dir: str, offline: bool = False):
        """
        Manages multiple models and their versions downloaded from civitai.com.

        :param root_dir: The root directory where the models will be managed.
        :param offline: If True, the manager operates in offline mode, using locally downloaded resources.
        """
        self.root_dir = root_dir

        os.makedirs(self.root_dir, exist_ok=True)
        self._f_lock = os.path.join(root_dir, '.filelock')
        self.lock = FileLock(self._f_lock)
        self._offline = offline

    def _model_path(self, model_name: str, model_id: int):
        return os.path.join(self.root_dir, f'{_soft_name_strip(model_name)}__{model_id}')

    def _list_local_models(self) -> Iterator[Tuple[str, int, str]]:
        for dir_ in os.listdir(self.root_dir):
            segs = dir_.split('__')
            if os.path.isdir(os.path.join(self.root_dir, dir_)) and len(segs) == 2:
                model_name, model_id = segs
                model_id = int(model_id)
                yield model_name, model_id, os.path.join(self.root_dir, dir_)

    def _find_online_model(self, model_name_or_id: Union[str, int]):
        model_data = find_model(model_name_or_id)
        model_id, model_name = model_data['id'], model_data['name']
        return model_name, model_id, self._model_path(model_name, model_id), model_data

    def _find_local_model(self, model_name_or_id: Union[str, int]):
        valid_models = []
        for model_name, model_id, model_dir in self._list_local_models():
            if (_soft_name_strip(str(model_name_or_id)) == model_name) or (model_id == model_name_or_id):
                valid_models.append((model_name, model_id, model_dir))

        if not valid_models:
            raise LocalModelNotFound(model_name_or_id)
        elif len(valid_models) > 1:
            raise LocalModelDuplicated(valid_models)
        else:
            model_name, model_id, model_dir = valid_models[0]
            return model_name, model_id, model_dir

    def _get_model_manager(self, model_name_or_id: Union[str, int]):
        try:
            if OFFLINE_MODE or self._offline:
                raise OfflineModeEnabled

            model_name, model_id, model_dir, model_data = self._find_online_model(model_name_or_id)
            return ModelManager(model_dir, model_name_or_id, model_data, offline=False)
        except (requests.exceptions.SSLError, requests.exceptions.ProxyError):
            # Actually raise for those subclasses of ConnectionError
            raise
        except (
                requests.exceptions.ConnectionError,
                requests.exceptions.Timeout,
                OfflineModeEnabled,
        ):
            model_name, model_id, model_dir = self._find_local_model(model_name_or_id)
            return ModelManager(model_dir, model_name_or_id, offline=True)

    def get_file(self, model_name_or_id: Union[str, int],
                 version: Union[str, int, None] = None, pattern: str = None):
        """
        Get the local file path of the specified model file.

        :param model_name_or_id: The name or ID of the model to manage.
        :type model_name_or_id: Union[str, int]
        :param version: The version ID or name of the model version. If None, the latest version is used.
        :type version: Union[str, int, None]
        :param pattern: The pattern or name of the file to get. If None, the primary file will be returned.
        :type pattern: str
        :return: The local path of the specified model file.
        :rtype: str
        :raises LocalModelNotFound: If the specified model is not found locally.
        :raises LocalModelDuplicated: If multiple models matching the model_name_or_id are found locally.
        """
        with self.lock:
            return self._get_model_manager(model_name_or_id).get_file(version, pattern)

    def list_models(self) -> List[ModelManager]:
        """
        List all the local models managed by this DispatchManager.

        :return: A list of ModelManager objects, one for each model.
        :rtype: List[ModelManager]
        """
        with self.lock:
            retval = []
            for model_name, model_id, model_dir in self._list_local_models():
                retval.append(ModelManager(model_dir, model_name, offline=True))

            return retval

    @property
    def total_size(self) -> int:
        """
        Get the total size of all the local model files managed by this DispatchManager.

        :return: The total size in bytes.
        :rtype: int
        """
        with self.lock:
            return sum((model.total_size for model in self.list_models()))

    def delete_model(self, model_name_or_id: Union[str, int]):
        """
        Delete the specified model and all its versions from the local storage.

        :param model_name_or_id: The name or ID of the model to delete.
        :type model_name_or_id: Union[str, int]
        :raises LocalModelNotFound: If the specified model is not found locally.
        """
        with self.lock:
            model_name, model_id, model_dir = self._find_local_model(model_name_or_id)
            shutil.rmtree(model_dir, ignore_errors=True)

    def delete_version(self, model_name_or_id: Union[str, int], version: Union[str, int]):
        """
        Delete the specified version of the model from the local storage.

        :param model_name_or_id: The name or ID of the model.
        :type model_name_or_id: Union[str, int]
        :param version: The version ID or name to delete.
        :type version: Union[str, int]
        """
        with self.lock:
            self._get_model_manager(model_name_or_id).delete_version(version)

    def _repr(self):
        return f'<{self.__class__.__name__} directory: {self.root_dir!r}>'

    def _tree(self):
        return self, [item._tree() for item in sorted(self.list_models(), key=repr)]

    def __str__(self):
        return format_tree(
            nested_map(repr, self._tree()),
            lambda x: x[0],
            lambda x: x[1],
        )

    def __repr__(self):
        return self._repr()



import fnmatch
import re
from dataclasses import dataclass
from typing import Union

from .http import get_session, ENDPOINT


class ModelNotFound(Exception):
    pass


class ModelFoundDuplicated(Exception):
    pass


class ModelVersionNotFound(Exception):
    pass


class ModelVersionDuplicated(Exception):
    pass


class ResourceNotFound(Exception):
    pass


class ResourceDuplicated(Exception):
    pass


@dataclass
class Resource:
    """
    Data class for resource.
    """
    model_name: str
    model_id: int
    version_name: str
    version_id: int
    filename: str
    is_primary: bool
    url: str
    sha256: str
    size: int


def find_model_by_id(model_id) -> dict:
    """
    Retrieve model information from the CiviTAI API based on the given model ID.

    :param model_id: The ID of the model to retrieve information for.
    :type model_id: int
    :return: The dictionary containing the model information.
    :rtype: dict
    :raises ModelNotFound: If the model with the given ID is not found.
    """
    resp = get_session().get(f'{ENDPOINT}/api/v1/models/{model_id}')
    if resp.ok:
        return resp.json()
    else:
        if resp.status_code == 404:
            raise ModelNotFound(model_id)
        else:
            resp.raise_for_status()


def _name_strip(name: str) -> str:
    return re.sub(r'[\W_]+', '', name.lower())


def find_model_by_name(model_name: str) -> dict:
    """
    Retrieve model information from the CiviTAI API based on the given model name.

    :param model_name: The name of the model to retrieve information for.
    :type model_name: str
    :return: The dictionary containing the model information.
    :rtype: dict
    :raises ModelNotFound: If the model with the given name is not found.
    :raises ModelFoundDuplicated: If multiple models with the same name are found.
    """
    resp = get_session().get(f'{ENDPOINT}/api/v1/models', params={'query': model_name})
    resp.raise_for_status()
    collected_items = []
    for item in resp.json()['items']:
        if _name_strip(item['name']) == _name_strip(model_name):
            collected_items.append(item)

    if not collected_items:
        raise ModelNotFound(model_name)
    elif len(collected_items) > 1:
        all_names = [item['name'] for item in collected_items]
        raise ModelFoundDuplicated(model_name, all_names)
    else:
        return collected_items[0]


def find_model(model_name_or_id: Union[int, str]) -> dict:
    """
    Find model information from the CiviTAI API based on the given model name or ID.

    :param model_name_or_id: The name or ID of the model to retrieve information for.
    :type model_name_or_id: Union[int, str]
    :return: The dictionary containing the model information.
    :rtype: dict
    :raises TypeError: If the model name or ID is not a valid integer or string.
    """
    if isinstance(model_name_or_id, int):
        return find_model_by_id(model_name_or_id)
    elif isinstance(model_name_or_id, str):
        try:
            model_id = int(model_name_or_id)
            return find_model_by_id(model_id)
        except (ModelNotFound, TypeError, ValueError):
            return find_model_by_name(model_name_or_id)
    else:
        raise TypeError(f'Unknown model name or id, it should be an integer or string - {model_name_or_id!r}.')


def find_version(model_data: dict, version: Union[int, str, None] = None):
    """
    Find the specified version from the model data.

    :param model_data: The model data containing version information.
    :type model_data: dict
    :param version: The version ID or name to find. If None, the first version will be chosen.
    :type version: Union[int, str, None]
    :return: The dictionary containing the selected version information.
    :rtype: dict
    :raises ModelVersionNotFound: If the specified version is not found.
    :raises ModelVersionDuplicated: If multiple versions with the same ID or name are found.
    """
    # find chosen version
    versions = model_data['modelVersions']
    if version is None:
        select_version = versions[0]
    else:
        all_select_versions = []
        for vitem in versions:
            if vitem['id'] == version or _name_strip(vitem['name']) == _name_strip(str(version)):
                all_select_versions.append(vitem)

        if not all_select_versions:
            raise ModelVersionNotFound(model_data['name'], version)
        elif len(all_select_versions) > 1:
            raise ModelVersionDuplicated(model_data['name'], [vitem['name'] for vitem in all_select_versions])
        else:
            select_version = all_select_versions[0]

    return select_version


def find_resource(model_data: dict, version_data: dict, pattern: str = None):
    """
    Find the specified resource from the model and version data.

    :param model_data: The model data containing version information.
    :type model_data: dict
    :param version_data: The version data containing resource information.
    :type version_data: dict
    :param pattern: The pattern to match the resource filename. If None, the primary resource will be chosen.
    :type pattern: str
    :return: The resource information.
    :rtype: Resource
    :raises TypeError: If the pattern of the resource is not a valid string.
    :raises ResourceNotFound: If the specified resource is not found.
    :raises ResourceDuplicated: If multiple resources with the same name are found.
    """
    # find chosen resource
    if pattern is None:
        pattern, primary = '*', True
    elif isinstance(pattern, str):
        primary = False
    else:
        raise TypeError(f'Pattern of resource should be a string, but {pattern!r} found.')

    all_select_files = []
    for file in version_data['files']:
        if (primary and file.get('primary')) or (not primary and fnmatch.fnmatch(file['name'], pattern)):
            all_select_files.append(file)
    if not all_select_files:
        raise ResourceNotFound(model_data['name'], version_data['name'], {'pattern': pattern, 'primary': primary})
    elif len(all_select_files) > 1:
        raise ResourceDuplicated(model_data['name'], version_data['name'], [file['name'] for file in all_select_files])
    else:
        select_file = all_select_files[0]

    file_size = select_file['sizeKB'] * 1024
    assert abs(round(file_size) - file_size) < 1e-4
    file_size = int(round(file_size))
    return Resource(
        model_name=model_data['name'], model_id=model_data['id'],
        version_name=version_data['name'], version_id=version_data['id'],
        filename=select_file['name'],
        url=select_file['downloadUrl'],
        sha256=select_file['hashes']['SHA256'],
        is_primary=select_file.get('primary', False),
        size=file_size,
    )



import os
from functools import lru_cache
from typing import Union

from .client import find_model, find_version, find_resource, Resource
from .manager import DispatchManager


@lru_cache()
def _get_storage_dir():
    """
    Get the directory path where the downloaded models will be managed.

    :return: The storage directory path.
    :rtype: str
    """
    return os.environ.get('CIVITAI_HOME', os.path.expanduser('~/.cache/civitai'))


@lru_cache()
def _get_global_manager(offline: bool):
    """
    Get the global manager responsible for managing all downloaded models.

    :param offline: If True, the manager operates in offline mode, using locally downloaded resources.
    :return: The global DispatchManager instance.
    :rtype: DispatchManager
    """
    return DispatchManager(_get_storage_dir(), offline)


def civitai_download(model: Union[str, int], version: Union[str, int, None] = None,
                     file: str = None, offline: bool = False):
    """
    Download and get the local file path of the specified model file.

    :param model: The name or ID of the model to download and manage.
    :type model: Union[str, int]
    :param version: The version ID or name of the model version. If None, the latest version is used.
    :type version: Union[str, int, None]
    :param file: The pattern or name of the file to get. If None, the primary file will be returned.
    :type file: str
    :param offline: If True, the manager operates in offline mode, using locally downloaded resources.
    :type offline: bool
    :return: The local path of the specified model file.
    :rtype: str
    """
    return _get_global_manager(offline).get_file(model, version, file)


def civitai_find_online(model: Union[str, int], version: Union[str, int, None] = None, file: str = None) -> Resource:
    """
    Find the online model resource (file) information from civitai.com.

    :param model: The name or ID of the model.
    :type model: Union[str, int]
    :param version: The version ID or name of the model version. If None, the latest version is used.
    :type version: Union[str, int, None]
    :param file: The pattern or name of the file to find. If None, the primary file will be returned.
    :type file: str
    :return: The Resource object containing the information about the specified model file.
    :rtype: Resource
    """
    model_data = find_model(model)
    version_data = find_version(model_data, version)
    return find_resource(model_data, version_data, file)



import hashlib
import os

import pytest

from pycivitai.client import ResourceDuplicated
from pycivitai.dispatch import civitai_find_online, civitai_download


def calculate_sha256(file_path):
    sha256_hash = hashlib.sha256()
    block_size = 65536

    with open(file_path, 'rb') as file:
        while True:
            data = file.read(block_size)
            if not data:
                break
            sha256_hash.update(data)

    return sha256_hash.hexdigest().upper()


@pytest.mark.unittest
class TestDispatch:
    def test_civitai_find_online(self):
        resource = civitai_find_online('明日方舟-安洁莉娜,Arknights-Angeline')
        assert resource.model_id == 5632
        assert resource.model_name == '明日方舟-安洁莉娜,Arknights-Angeline'
        assert resource.version_id == 6555
        assert resource.version_name == 'v1.0'
        assert resource.filename == '安洁莉娜5(头巾headband,三套衣服coat,multiple strap,swimsuit),.safetensors'
        assert resource.sha256 == '89A4D7CF20C76AFEE5AA4C4F28908A05D5E6BB6EB7726E5D319DE74731D09701'
        assert resource.is_primary
        assert resource.size == 37863532

        resource = civitai_find_online('Cetus-Mix', 'V4')
        assert resource.model_id == 6755
        assert resource.model_name == 'Cetus-Mix'
        assert resource.version_id == 78676
        assert resource.version_name == 'V4'
        assert resource.filename == 'cetusMix_v4.safetensors'
        assert resource.sha256 == 'B42B09FF12CA9CD70D78AA8210F8D4577EC513FC1484A68615385B8076292639'
        assert resource.is_primary
        assert resource.size == 3894258133

        resource = civitai_find_online('Cetus-Mix', 'V4', '*.safetensors')
        assert resource.model_id == 6755
        assert resource.model_name == 'Cetus-Mix'
        assert resource.version_id == 78676
        assert resource.version_name == 'V4'
        assert resource.filename == 'cetusMix_v4.safetensors'
        assert resource.sha256 == 'B42B09FF12CA9CD70D78AA8210F8D4577EC513FC1484A68615385B8076292639'
        assert resource.is_primary
        assert resource.size == 3894258133

        resource = civitai_find_online('Cetus-Mix', 'V4', '*.vae.pt')
        assert resource.model_id == 6755
        assert resource.model_name == 'Cetus-Mix'
        assert resource.version_id == 78676
        assert resource.version_name == 'V4'
        assert resource.filename == 'MoistMix.vae.pt'
        assert resource.sha256 == 'DF3C506E51B7EE1D7B5A6A2BB7142D47D488743C96AA778AFB0F53A2CDC2D38D'
        assert not resource.is_primary
        assert resource.size == 404662241

        with pytest.raises(ResourceDuplicated):
            _ = civitai_find_online('Cetus-Mix', 'V4', '*')

    def test_civitai_download(self):
        file = civitai_download('amiya arknights (old)')
        assert os.path.getsize(file) == 25515
        assert calculate_sha256(file) == '311279B35743D703DED68DF6ECC9F1E850A9F6976494CC8AF434DF3AA0E238A0'

        file = civitai_download('明日方舟-安洁莉娜,Arknights-Angeline')
        assert os.path.getsize(file) == 37863532
        assert calculate_sha256(file) == '89A4D7CF20C76AFEE5AA4C4F28908A05D5E6BB6EB7726E5D319DE74731D09701'



import fnmatch
import logging
import os.path
import pathlib
import shutil
from dataclasses import dataclass
from typing import Union, Optional, Tuple, Iterator, List

import requests
from filelock import FileLock
from hbutils.collection import nested_map
from hbutils.string import format_tree
from hbutils.system import TemporaryDirectory

from ..client import get_session, ENDPOINT, find_resource, Resource, find_model, find_version, OFFLINE_MODE, \
    OfflineModeEnabled
from ..utils import download_file


class LocalPrimaryFileUnset(Exception):
    pass


class LocalFileNotFound(Exception):
    pass


class LocalFileDuplicated(Exception):
    pass


@dataclass
class LocalFile:
    """
    Data class of local managed file.
    """
    filename: str
    hash: str
    size: int
    is_primary: bool


class VersionManager:
    """
    Management of specific version of model.
    """

    def __init__(self, root_dir: str, model_name_or_id: Union[str, int], version: Union[str, int],
                 model_data: Optional[dict] = None, offline: bool = False):
        """
        Manages the local model files downloaded from civitai.com for a specific model and version.

        :param root_dir: The root directory where the model files will be managed.
        :param model_name_or_id: The name or ID of the model to manage files for.
        :param version: The version ID or name to manage files for.
        :param model_data: Optional dictionary containing model information to avoid fetching it from the API.
        :param offline: If True, the manager operates in offline mode, using locally downloaded resources.
        """
        self.root_dir = root_dir
        self.model_name_or_id = model_name_or_id
        self._model_data = model_data
        self.version = version
        self._version_data = None

        os.makedirs(self.root_dir, exist_ok=True)
        self._f_lock = os.path.join(self.root_dir, '.filelock')
        self._f_primary = os.path.join(self.root_dir, 'primary')
        self._d_files = os.path.join(self.root_dir, 'files')
        self._d_hashes = os.path.join(self.root_dir, 'hashes')
        self.lock = FileLock(self._f_lock)
        self._offline = offline

    @property
    def _primary_file(self) -> Optional[str]:
        if os.path.exists(self._f_primary):
            pfile_val = pathlib.Path(self._f_primary).read_text(encoding='utf-8').splitlines(keepends=False)[0]
            return pfile_val if pfile_val else None
        else:
            return None

    def _get_model(self):
        if not self._model_data:
            self._model_data = find_model(self.model_name_or_id)
        return self._model_data

    def _get_version(self):
        if not self._version_data:
            self._version_data = find_version(self._get_model(), self.version)
        return self._version_data

    def _get_resource(self, pattern: str = None) -> Resource:
        return find_resource(self._get_model(), self._get_version(), pattern)

    def _file_path(self, filename: str):
        return os.path.join(self._d_files, filename)

    def _get_file_size(self, filename: str) -> Optional[int]:
        f = self._file_path(filename)
        return os.path.getsize(f) if os.path.exists(f) else None

    def _hash_path(self, filename: str):
        return os.path.join(self._d_hashes, f'{filename}.hash')

    def _get_file_hash(self, filename: str) -> Optional[str]:
        f = self._hash_path(filename)
        return pathlib.Path(f).read_text(encoding='utf-8').strip() if os.path.exists(f) else None

    def _get_file_meta(self, filename: str) -> Tuple[Optional[str], Optional[int]]:
        return self._get_file_hash(filename), self._get_file_size(filename)

    def _need_download_check(self, resource: Resource):
        local_hash, local_size = self._get_file_meta(resource.filename)
        if not local_hash or local_size is None:
            return True
        else:
            return (resource.size != local_size) or (resource.sha256 != local_hash)

    def _iter_local_files(self) -> Iterator[Tuple[str, str, int]]:
        if os.path.exists(self._d_files):
            for f in os.listdir(self._d_files):
                _hash, _size = self._get_file_meta(f)
                if not _hash or _size is None:
                    continue

                yield f, _hash, _size

    def _download_resource(self, resource: Resource):
        with TemporaryDirectory() as td:
            file = os.path.join(td, resource.filename)
            download_file(
                resource.url, file,
                expected_size=resource.size,
                desc=resource.filename,
                session=get_session(),
            )

            os.makedirs(self._d_files, exist_ok=True)
            shutil.copyfile(file, self._file_path(resource.filename))
            os.makedirs(self._d_hashes, exist_ok=True)
            with open(self._hash_path(resource.filename), 'w', encoding='utf-8') as hf:
                hf.write(resource.sha256)

            if resource.is_primary:
                with open(self._f_primary, 'w', encoding='utf-8') as pf:
                    pf.write(resource.filename)

    def _try_sync_from_site(self, pattern: str = None):
        try:
            if OFFLINE_MODE or self._offline:
                raise OfflineModeEnabled

            resource = self._get_resource(pattern)
            logging.debug(f'Resource found from {ENDPOINT!r}: {resource!r}')
            if self._need_download_check(resource):
                logging.debug('The resource is not available locally or has been updated. '
                              'The download will commence shortly.')
                self._download_resource(resource)
        except (requests.exceptions.SSLError, requests.exceptions.ProxyError):
            # Actually raise for those subclasses of ConnectionError
            raise
        except (
                requests.exceptions.ConnectionError,
                requests.exceptions.Timeout,
                OfflineModeEnabled,
        ):
            # ignore this, use the local models
            logging.debug('Offline environment detected, using locally downloaded resources.')

    def get_file(self, pattern: str = None):
        """
        Get the local file path of the specified model file.

        :param pattern: The pattern or name of the file to get. If None, the primary file will be returned.
        :type pattern: str
        :return: The local path of the specified model file.
        :rtype: str
        :raises LocalPrimaryFileUnset: If the primary file is not set and no specific pattern is provided.
        :raises LocalFileNotFound: If the specified file is not found locally.
        :raises LocalFileDuplicated: If multiple files matching the pattern are found locally.
        """
        with self.lock:
            self._try_sync_from_site(pattern)

            if pattern is None:
                pattern = self._primary_file
                fullmatch = True
                if pattern is None:
                    raise LocalPrimaryFileUnset(self.model_name_or_id, self.version)
            else:
                fullmatch = False

            matched_files = []
            if os.path.exists(self._d_files):
                for filename in os.listdir(self._d_files):
                    if (fullmatch and filename == pattern) or \
                            (not fullmatch and fnmatch.fnmatch(filename, pattern)):
                        matched_files.append(filename)

            if not matched_files:
                raise LocalFileNotFound(self.model_name_or_id, self.version)
            elif len(matched_files) > 1:
                raise LocalFileDuplicated(self.model_name_or_id, self.version, matched_files)
            else:
                local_file = self._file_path(matched_files[0])
                assert os.path.exists(local_file), \
                    f'The expected resource file {local_file!r} was not found, ' \
                    f'indicating a BUG. Please contact the developer.'
                return local_file

    def list_files(self) -> List[LocalFile]:
        """
        List all the local model files associated with this version manager.

        :return: A list of LocalFile objects containing information about each local model file.
        :rtype: List[LocalFile]
        """
        with self.lock:
            retval = []
            primary_file = self._primary_file
            for filename, hash_, size_ in self._iter_local_files():
                retval.append(LocalFile(filename, hash_, size_, filename == primary_file))

            return retval

    @property
    def total_size(self) -> int:
        """
        Get the total size of all the local model files associated with this version manager.

        :return: The total size in bytes.
        :rtype: int
        """
        with self.lock:
            return sum((file.size for file in self.list_files()))

    def delete_file(self, filename):
        """
        Delete the specified model file from the local storage.

        :param filename: The name of the file to delete.
        :type filename: str
        :raises LocalFileNotFound: If the specified file is not found locally.
        """
        with self.lock:
            fp = self._file_path(filename)
            hp = self._hash_path(filename)
            if not os.path.exists(fp) and not os.path.exists(hp):
                raise LocalFileNotFound(self.model_name_or_id, self.version, filename)
            else:
                if os.path.exists(fp):
                    os.remove(fp)
                if os.path.exists(hp):
                    os.remove(hp)

    def _repr(self):
        return f'<{self.__class__.__name__} model: {self.model_name_or_id!r}, version: {self.version!r}>'

    def _tree(self):
        return self, [(item, []) for item in sorted(self.list_files(), key=repr)]

    def __str__(self):
        return format_tree(
            nested_map(repr, self._tree()),
            lambda x: x[0],
            lambda x: x[1],
        )

    def __repr__(self):
        return self._repr()



import logging
import os
from functools import partial

import click
from hbutils.scale import size_to_bytes_str
from hbutils.string import plural_word

from .dispatch import _get_global_manager, civitai_download
from .utils import GLOBAL_CONTEXT_SETTINGS
from .utils import print_version as _origin_print_version

try:
    import InquirerPy
except (ModuleNotFoundError, ImportError):
    InquirerPy = None


def _check_inquirerpy():
    if InquirerPy is None:
        raise OSError(
            'TUI for models deletion not available, '
            'please install CLI extra with `pip install pycivitiai[cli]`.'
        )


print_version = partial(_origin_print_version, 'pycivitai')


@click.group(context_settings={**GLOBAL_CONTEXT_SETTINGS}, help='Command-Line Interface for pycivitai')
@click.option('-v', '--version', is_flag=True,
              callback=print_version, expose_value=False, is_eager=True)
def cli():
    pass  # pragma: no cover


def _get_choices(choices):  # pragma: no cover
    from InquirerPy import inquirer
    return inquirer.checkbox(
        message='Choose model versions to delete:',
        choices=choices,
        cycle=False,
        transformer=lambda result: "%s region%s selected"
                                   % (len(result), "s" if len(result) > 1 else ""),
    ).execute()


def _confirm(text):  # pragma: no cover
    from InquirerPy import inquirer
    return inquirer.confirm(text).execute()


@cli.command('delete-cache', context_settings={**GLOBAL_CONTEXT_SETTINGS},
             help='Delete downloaded models from storage.')
@click.option('-A', '--all', 'delete_all', is_flag=True, type=bool, default=False,
              help='Delete all the downloaded models.', show_default=True)
def delete_cache(delete_all):
    manager = _get_global_manager(offline=True)
    with manager.lock:
        if delete_all:
            total_size = 0
            for model in manager.list_models():
                model_size = model.total_size
                logging.debug(f'Deleting {model.model_name_or_id!r}'
                              f'({size_to_bytes_str(model_size, precision=3)}) ...')
                manager.delete_model(model.model_name_or_id)
                total_size += model_size

            click.echo(f'All models deleted, total {size_to_bytes_str(total_size, precision=3)}.')

        else:
            _check_inquirerpy()

            from InquirerPy import inquirer
            from InquirerPy.base import Choice
            from InquirerPy.separator import Separator

            choices = []
            for model in manager.list_models():
                model_dir = model.root_dir
                model_name, model_id = os.path.basename(model_dir).split('__')

                model_total_size, model_files = 0, 0
                model_choices = []
                for version in model.list_versions():
                    version_dir = version.root_dir
                    version_name, version_id = os.path.basename(version_dir).split('__')
                    files = version.list_files()
                    total_size = sum([file.size for file in files])

                    if len(files) > 0:
                        model_files += len(files)
                        model_total_size += total_size
                        model_choices.append(Choice(
                            (model.model_name_or_id, version.version, len(files), total_size),
                            name=f'{version_name}(ID: {version_id}, {plural_word(len(files), "file")}, '
                                 f'size: {size_to_bytes_str(total_size, precision=3)})',
                            enabled=False
                        ))

                if model_choices:
                    choices.extend([
                        Separator(f'Model {model_name}(ID: {model_id}, {plural_word(model_files, "file")}, '
                                  f'size: {size_to_bytes_str(model_total_size, precision=3)}):'),
                        *model_choices,
                        Separator('')
                    ])

            if choices:
                versions_to_detect = _get_choices(choices)
                files_to_delete = sum((x[2] for x in versions_to_detect))
                size_to_delete = sum((x[3] for x in versions_to_detect))
                if files_to_delete == 0:
                    click.echo(click.style('Deletion cancelled.', fg='yellow'))
                elif _confirm(f'{plural_word(len(versions_to_detect), "version")} with '
                              f'{plural_word(files_to_delete, "file")} will be deleted, '
                              f'{size_to_bytes_str(size_to_delete, precision=3)} of disk usage '
                              f'will be released, confirm?'):
                    for model_name, version_name, _, _ in versions_to_detect:
                        manager.delete_version(model_name, version_name)
                    click.echo(click.style('Deletion complete!', fg='green'))

            else:
                click.echo(click.style('No models found to delete.', fg='yellow'))


@cli.command('get', context_settings={**GLOBAL_CONTEXT_SETTINGS},
             help='Delete downloaded models from storage.')
@click.option('--model', '-m', 'model', type=str, required=True,
              help='Title or id of the model.')
@click.option('--version', '-v', 'version', type=str, default=None,
              help='Name or id of the version. Latest version will be used when not given.', show_default=True)
@click.option('--file', '-f', 'file', type=str, default=None,
              help='Name pattern of the model file. Primary file will be used when not given', show_default=True)
@click.option('--offline', 'offline', is_flag=True, type=bool, default=False,
              help='Offline mode. Default is disabled.', show_default=True)
def get_(model, version, file, offline):
    click.echo(civitai_download(model, version, file, offline))



import types

import click
from click.core import Context, Option

from ..config.meta import __TITLE__, __VERSION__, __AUTHOR__, __AUTHOR_EMAIL__

_raw_authors = [item.strip() for item in __AUTHOR__.split(',') if item.strip()]
_raw_emails = [item.strip() for item in __AUTHOR_EMAIL__.split(',')]
if len(_raw_emails) < len(_raw_authors):  # pragma: no cover
    _raw_emails += [None] * (len(_raw_authors) - len(_raw_emails))
elif len(_raw_emails) > len(_raw_authors):  # pragma: no cover
    _raw_emails[len(_raw_authors) - 1] = tuple(_raw_emails[len(_raw_authors) - 1:])
    del _raw_emails[len(_raw_authors):]

_author_tuples = [
    (author, tuple([item for item in (email if isinstance(email, tuple) else ((email,) if email else ())) if item]))
    for author, email in zip(_raw_authors, _raw_emails)
]
_authors = [
    author if not emails else '{author} ({emails})'.format(author=author, emails=', '.join(emails))
    for author, emails in _author_tuples
]

GLOBAL_CONTEXT_SETTINGS = dict(
    help_option_names=['-h', '--help']
)


def print_version(module, ctx: Context, param: Option, value: bool) -> None:
    """
    Print version information of cli
    :param module: current module using this cli.
    :param ctx: click context
    :param param: current parameter's metadata
    :param value: value of current parameter
    """
    if not value or ctx.resilient_parsing:
        return  # pragma: no cover

    if module is None:
        title = __TITLE__
    elif isinstance(module, types.ModuleType):
        title = module.__name__
    else:
        title = str(module)

    click.echo('{title}, version {version}.'.format(title=title.capitalize(), version=__VERSION__))
    if _authors:
        click.echo('Developed by {authors}.'.format(authors=', '.join(_authors)))
    ctx.exit()



import os
import shutil
from typing import Union, Optional, Iterator, Tuple, List

import requests
from filelock import FileLock
from hbutils.collection import nested_map
from hbutils.string import format_tree

from .base import _soft_name_strip
from .version import VersionManager
from ..client import find_model, find_version, OFFLINE_MODE, OfflineModeEnabled


class LocalVersionNotFound(Exception):
    pass


class LocalVersionDuplicated(Exception):
    pass


class ModelManager:
    """
    Management of specific model.
    """

    def __init__(self, root_dir: str, model_name_or_id: Union[str, int],
                 model_data: Optional[dict] = None, offline: bool = False):
        """
        Manages multiple versions of a model downloaded from civitai.com.

        :param root_dir: The root directory where the model versions will be managed.
        :param model_name_or_id: The name or ID of the model to manage.
        :param model_data: Optional dictionary containing model information to avoid fetching it from the API.
        :param offline: If True, the manager operates in offline mode, using locally downloaded resources.
        """
        self.root_dir = root_dir
        self.model_name_or_id = model_name_or_id
        self._model_data = model_data

        os.makedirs(root_dir, exist_ok=True)
        self._d_versions = os.path.join(self.root_dir)
        self._f_lock = os.path.join(self.root_dir, '.filelock')
        self.lock = FileLock(self._f_lock)
        self._offline = offline

    def _get_model(self):
        if not self._model_data:
            self._model_data = find_model(self.model_name_or_id)
        return self._model_data

    def _version_path(self, version_name: str, version_id: int):
        return os.path.join(self._d_versions, f'{_soft_name_strip(version_name)}__{version_id}')

    def _list_local_versions(self) -> Iterator[Tuple[str, int, str]]:
        for dir_ in os.listdir(self._d_versions):
            segs = dir_.split('__')
            if os.path.isdir(os.path.join(self._d_versions, dir_)) and len(segs) == 2:
                version_name, version_id = segs
                version_id = int(version_id)
                yield version_name, version_id, os.path.join(self._d_versions, dir_)

    def _find_online_version(self, version: Union[str, int, None]):
        version_data = find_version(self._get_model(), version)
        version_id, version_name = version_data['id'], version_data['name']
        return version_name, version_id, self._version_path(version_name, version_id)

    def _find_local_version(self, version: Union[str, int, None]):
        valid_versions = []
        for version_name, version_id, version_dir in self._list_local_versions():
            if (version is None) or ((version is not None) and (
                    (_soft_name_strip(str(version)) == version_name) or
                    (version_id == version)
            )):
                valid_versions.append((version_name, version_id, version_dir))

        if not valid_versions:
            raise LocalVersionNotFound(self.model_name_or_id, version)
        else:
            if version is None:
                version_name, version_id, version_dir = sorted(valid_versions, key=lambda x: -x[1])[0]
            else:
                if len(valid_versions) > 1:
                    raise LocalVersionDuplicated(self.model_name_or_id, valid_versions)
                else:
                    version_name, version_id, version_dir = valid_versions[0]

            return version_name, version_id, version_dir

    def _get_version_manager(self, version: Union[str, int, None]):
        try:
            if OFFLINE_MODE or self._offline:
                raise OfflineModeEnabled

            version_name, version_id, version_dir = self._find_online_version(version)
            return VersionManager(
                version_dir, self.model_name_or_id, version_name,
                model_data=self._get_model(), offline=False
            )

        except (requests.exceptions.SSLError, requests.exceptions.ProxyError):
            # Actually raise for those subclasses of ConnectionError
            raise
        except (
                requests.exceptions.ConnectionError,
                requests.exceptions.Timeout,
                OfflineModeEnabled,
        ):
            version_name, version_id, version_dir = self._find_local_version(version)
            return VersionManager(version_dir, self.model_name_or_id, version_name, offline=True)

    def get_file(self, version: Union[str, int, None] = None, pattern: str = None):
        """
        Get the local file path of the specified model file.

        :param version: The version ID or name of the model version. If None, the latest version is used.
        :type version: Union[str, int, None]
        :param pattern: The pattern or name of the file to get. If None, the primary file will be returned.
        :type pattern: str
        :return: The local path of the specified model file.
        :rtype: str
        :raises LocalVersionNotFound: If the specified version is not found locally.
        :raises LocalVersionDuplicated: If multiple versions matching the version parameter are found locally.
        """
        with self.lock:
            return self._get_version_manager(version).get_file(pattern)

    def list_versions(self) -> List[VersionManager]:
        """
        List all the local model versions managed by this ModelManager.

        :return: A list of VersionManager objects, one for each version.
        :rtype: List[VersionManager]
        """
        with self.lock:
            retval = []
            for version_name, version_id, version_dir in self._list_local_versions():
                retval.append(VersionManager(version_dir, self.model_name_or_id, version_name, offline=True))

            return retval

    @property
    def total_size(self) -> int:
        """
        Get the total size of all the local model files managed by this ModelManager.

        :return: The total size in bytes.
        :rtype: int
        """
        return sum((version.total_size for version in self.list_versions()))

    def delete_version(self, version: Union[str, int, None] = None):
        """
        Delete the specified model version from the local storage.

        :param version: The version ID or name to delete. If None, the latest version is deleted.
        :type version: Union[str, int, None]
        :raises LocalVersionNotFound: If the specified version is not found locally.
        """
        with self.lock:
            version_name, version_id, version_dir = self._find_local_version(version)
            shutil.rmtree(version_dir, ignore_errors=True)

    def _tree(self):
        return self, [item._tree() for item in sorted(self.list_versions(), key=repr)]

    def _repr(self):
        return f'<{self.__class__.__name__} model: {self.model_name_or_id!r}>'

    def __str__(self):
        return format_tree(
            nested_map(repr, self._tree()),
            lambda x: x[0],
            lambda x: x[1],
        )

    def __repr__(self):
        return self._repr()
